HND,Tokyo Haneda International Airport,TOKYO,Asia/Tokyo,AS,JP,JP-13,t,
NRT,Narita International Airport,TOKYO,Asia/Tokyo,AS,JP,JP-12,t,
DEL,Indira Gandhi International Airport,DELHI,Asia/Calcutta,AS,IN,IN-DL,t,
SHA,Shanghai Hongqiao International Airport,SHANGHAI,Asia/Shanghai,AS,CN,CN-31,t,
CGH,Congonhas Airport,SAO PAULO,America/Sao_Paulo,SA,BR,BR-SP,f,
GRU,Guarulhos - Governador AndrÃ© Franco Montoro International Airport,SAO PAULO,America/Sao_Paulo,SA,BR,BR-SP,t,
MEX,Licenciado Benito Juarez International Airport,MEXICO CITY,America/Mexico_City,NA,MX,MX-DIF,t,
DAC,Hazrat Shahjalal International Airport,DHAKA,Asia/Dhaka,AS,BD,BD-3,t,
CAI,Cairo International Airport,CAIRO,Africa/Cairo,AF,EG,EG-C,t,
PEK,Beijing Capital International Airport,BEIJING,Asia/Shanghai,AS,CN,CN-11,t,