select ai.sc_set_table_desc(42, 0, 'postgres_air', 'account', 'this is a description for table account', 'my_catalog');
select ai.sc_set_table_col_desc(42, 0, 1, 'postgres_air', 'account', 'account_id', 'this is a description for column account_id', 'my_catalog');
select ai.sc_set_table_col_desc(42, 0, 2, 'postgres_air', 'account', 'login', 'this is a description for column login', 'my_catalog');
select ai.sc_set_table_col_desc(42, 0, 3, 'postgres_air', 'account', 'first_name', 'this is a description for column first_name', 'my_catalog');
select ai.sc_set_table_col_desc(42, 0, 4, 'postgres_air', 'account', 'last_name', 'this is a description for column last_name', 'my_catalog');
select ai.sc_set_table_col_desc(42, 0, 5, 'postgres_air', 'account', 'frequent_flyer_id', 'this is a description for column frequent_flyer_id', 'my_catalog');
select ai.sc_set_table_col_desc(42, 0, 6, 'postgres_air', 'account', 'update_ts', 'this is a description for column update_ts', 'my_catalog');
select ai.sc_set_table_desc(42, 1, 'postgres_air', 'aircraft', 'this is a description for table aircraft', 'my_catalog');
select ai.sc_set_table_col_desc(42, 1, 1, 'postgres_air', 'aircraft', 'model', 'this is a description for column model', 'my_catalog');
select ai.sc_set_table_col_desc(42, 1, 2, 'postgres_air', 'aircraft', 'range', 'this is a description for column range', 'my_catalog');
select ai.sc_set_table_col_desc(42, 1, 3, 'postgres_air', 'aircraft', 'class', 'this is a description for column class', 'my_catalog');
select ai.sc_set_table_col_desc(42, 1, 4, 'postgres_air', 'aircraft', 'velocity', 'this is a description for column velocity', 'my_catalog');
select ai.sc_set_table_col_desc(42, 1, 5, 'postgres_air', 'aircraft', 'code', 'this is a description for column code', 'my_catalog');
select ai.sc_set_table_desc(42, 2, 'postgres_air', 'airport', 'this is a description for table airport', 'my_catalog');
select ai.sc_set_table_col_desc(42, 2, 1, 'postgres_air', 'airport', 'airport_code', 'this is a description for column airport_code', 'my_catalog');
select ai.sc_set_table_col_desc(42, 2, 2, 'postgres_air', 'airport', 'airport_name', 'this is a description for column airport_name', 'my_catalog');
select ai.sc_set_table_col_desc(42, 2, 3, 'postgres_air', 'airport', 'city', 'this is a description for column city', 'my_catalog');
select ai.sc_set_table_col_desc(42, 2, 4, 'postgres_air', 'airport', 'airport_tz', 'this is a description for column airport_tz', 'my_catalog');
select ai.sc_set_table_col_desc(42, 2, 5, 'postgres_air', 'airport', 'continent', 'this is a description for column continent', 'my_catalog');
select ai.sc_set_table_col_desc(42, 2, 6, 'postgres_air', 'airport', 'iso_country', 'this is a description for column iso_country', 'my_catalog');
select ai.sc_set_table_col_desc(42, 2, 7, 'postgres_air', 'airport', 'iso_region', 'this is a description for column iso_region', 'my_catalog');
select ai.sc_set_table_col_desc(42, 2, 8, 'postgres_air', 'airport', 'intnl', 'this is a description for column intnl', 'my_catalog');
select ai.sc_set_table_col_desc(42, 2, 9, 'postgres_air', 'airport', 'update_ts', 'this is a description for column update_ts', 'my_catalog');
select ai.sc_set_table_desc(42, 3, 'postgres_air', 'boarding_pass', 'this is a description for table boarding_pass', 'my_catalog');
select ai.sc_set_table_col_desc(42, 3, 1, 'postgres_air', 'boarding_pass', 'pass_id', 'this is a description for column pass_id', 'my_catalog');
select ai.sc_set_table_col_desc(42, 3, 2, 'postgres_air', 'boarding_pass', 'passenger_id', 'this is a description for column passenger_id', 'my_catalog');
select ai.sc_set_table_col_desc(42, 3, 3, 'postgres_air', 'boarding_pass', 'booking_leg_id', 'this is a description for column booking_leg_id', 'my_catalog');
select ai.sc_set_table_col_desc(42, 3, 4, 'postgres_air', 'boarding_pass', 'seat', 'this is a description for column seat', 'my_catalog');
select ai.sc_set_table_col_desc(42, 3, 5, 'postgres_air', 'boarding_pass', 'boarding_time', 'this is a description for column boarding_time', 'my_catalog');
select ai.sc_set_table_col_desc(42, 3, 6, 'postgres_air', 'boarding_pass', 'precheck', 'this is a description for column precheck', 'my_catalog');
select ai.sc_set_table_col_desc(42, 3, 7, 'postgres_air', 'boarding_pass', 'update_ts', 'this is a description for column update_ts', 'my_catalog');
select ai.sc_set_table_desc(42, 4, 'postgres_air', 'booking', 'this is a description for table booking', 'my_catalog');
select ai.sc_set_table_col_desc(42, 4, 1, 'postgres_air', 'booking', 'booking_id', 'this is a description for column booking_id', 'my_catalog');
select ai.sc_set_table_col_desc(42, 4, 2, 'postgres_air', 'booking', 'booking_ref', 'this is a description for column booking_ref', 'my_catalog');
select ai.sc_set_table_col_desc(42, 4, 3, 'postgres_air', 'booking', 'booking_name', 'this is a description for column booking_name', 'my_catalog');
select ai.sc_set_table_col_desc(42, 4, 4, 'postgres_air', 'booking', 'account_id', 'this is a description for column account_id', 'my_catalog');
select ai.sc_set_table_col_desc(42, 4, 5, 'postgres_air', 'booking', 'email', 'this is a description for column email', 'my_catalog');
select ai.sc_set_table_col_desc(42, 4, 6, 'postgres_air', 'booking', 'phone', 'this is a description for column phone', 'my_catalog');
select ai.sc_set_table_col_desc(42, 4, 7, 'postgres_air', 'booking', 'update_ts', 'this is a description for column update_ts', 'my_catalog');
select ai.sc_set_table_col_desc(42, 4, 8, 'postgres_air', 'booking', 'price', 'this is a description for column price', 'my_catalog');
select ai.sc_set_table_desc(42, 5, 'postgres_air', 'booking_leg', 'this is a description for table booking_leg', 'my_catalog');
select ai.sc_set_table_col_desc(42, 5, 1, 'postgres_air', 'booking_leg', 'booking_leg_id', 'this is a description for column booking_leg_id', 'my_catalog');
select ai.sc_set_table_col_desc(42, 5, 2, 'postgres_air', 'booking_leg', 'booking_id', 'this is a description for column booking_id', 'my_catalog');
select ai.sc_set_table_col_desc(42, 5, 3, 'postgres_air', 'booking_leg', 'flight_id', 'this is a description for column flight_id', 'my_catalog');
select ai.sc_set_table_col_desc(42, 5, 4, 'postgres_air', 'booking_leg', 'leg_num', 'this is a description for column leg_num', 'my_catalog');
select ai.sc_set_table_col_desc(42, 5, 5, 'postgres_air', 'booking_leg', 'is_returning', 'this is a description for column is_returning', 'my_catalog');
select ai.sc_set_table_col_desc(42, 5, 6, 'postgres_air', 'booking_leg', 'update_ts', 'this is a description for column update_ts', 'my_catalog');
select ai.sc_set_table_desc(42, 6, 'postgres_air', 'events', 'this is a description for table events', 'my_catalog');
select ai.sc_set_table_col_desc(42, 6, 1, 'postgres_air', 'events', 'time', 'this is a description for column time', 'my_catalog');
select ai.sc_set_table_col_desc(42, 6, 2, 'postgres_air', 'events', 'name', 'this is a description for column name', 'my_catalog');
select ai.sc_set_table_col_desc(42, 6, 3, 'postgres_air', 'events', 'params', 'this is a description for column params', 'my_catalog');
select ai.sc_set_table_desc(42, 7, 'postgres_air', 'flight', 'this is a description for table flight', 'my_catalog');
select ai.sc_set_table_col_desc(42, 7, 1, 'postgres_air', 'flight', 'flight_id', 'this is a description for column flight_id', 'my_catalog');
select ai.sc_set_table_col_desc(42, 7, 2, 'postgres_air', 'flight', 'flight_no', 'this is a description for column flight_no', 'my_catalog');
select ai.sc_set_table_col_desc(42, 7, 3, 'postgres_air', 'flight', 'scheduled_departure', 'this is a description for column scheduled_departure', 'my_catalog');
select ai.sc_set_table_col_desc(42, 7, 4, 'postgres_air', 'flight', 'scheduled_arrival', 'this is a description for column scheduled_arrival', 'my_catalog');
select ai.sc_set_table_col_desc(42, 7, 5, 'postgres_air', 'flight', 'departure_airport', 'this is a description for column departure_airport', 'my_catalog');
select ai.sc_set_table_col_desc(42, 7, 6, 'postgres_air', 'flight', 'arrival_airport', 'this is a description for column arrival_airport', 'my_catalog');
select ai.sc_set_table_col_desc(42, 7, 7, 'postgres_air', 'flight', 'status', 'this is a description for column status', 'my_catalog');
select ai.sc_set_table_col_desc(42, 7, 8, 'postgres_air', 'flight', 'aircraft_code', 'this is a description for column aircraft_code', 'my_catalog');
select ai.sc_set_table_col_desc(42, 7, 9, 'postgres_air', 'flight', 'actual_departure', 'this is a description for column actual_departure', 'my_catalog');
select ai.sc_set_table_col_desc(42, 7, 10, 'postgres_air', 'flight', 'actual_arrival', 'this is a description for column actual_arrival', 'my_catalog');
select ai.sc_set_table_col_desc(42, 7, 11, 'postgres_air', 'flight', 'update_ts', 'this is a description for column update_ts', 'my_catalog');
select ai.sc_set_table_desc(42, 8, 'postgres_air', 'frequent_flyer', 'this is a description for table frequent_flyer', 'my_catalog');
select ai.sc_set_table_col_desc(42, 8, 1, 'postgres_air', 'frequent_flyer', 'frequent_flyer_id', 'this is a description for column frequent_flyer_id', 'my_catalog');
select ai.sc_set_table_col_desc(42, 8, 2, 'postgres_air', 'frequent_flyer', 'first_name', 'this is a description for column first_name', 'my_catalog');
select ai.sc_set_table_col_desc(42, 8, 3, 'postgres_air', 'frequent_flyer', 'last_name', 'this is a description for column last_name', 'my_catalog');
select ai.sc_set_table_col_desc(42, 8, 4, 'postgres_air', 'frequent_flyer', 'title', 'this is a description for column title', 'my_catalog');
select ai.sc_set_table_col_desc(42, 8, 5, 'postgres_air', 'frequent_flyer', 'card_num', 'this is a description for column card_num', 'my_catalog');
select ai.sc_set_table_col_desc(42, 8, 6, 'postgres_air', 'frequent_flyer', 'level', 'this is a description for column level', 'my_catalog');
select ai.sc_set_table_col_desc(42, 8, 7, 'postgres_air', 'frequent_flyer', 'award_points', 'this is a description for column award_points', 'my_catalog');
select ai.sc_set_table_col_desc(42, 8, 8, 'postgres_air', 'frequent_flyer', 'email', 'this is a description for column email', 'my_catalog');
select ai.sc_set_table_col_desc(42, 8, 9, 'postgres_air', 'frequent_flyer', 'phone', 'this is a description for column phone', 'my_catalog');
select ai.sc_set_table_col_desc(42, 8, 10, 'postgres_air', 'frequent_flyer', 'update_ts', 'this is a description for column update_ts', 'my_catalog');
select ai.sc_set_table_desc(42, 9, 'postgres_air', 'hypertable_test', 'this is a description for table hypertable_test', 'my_catalog');
select ai.sc_set_table_col_desc(42, 9, 1, 'postgres_air', 'hypertable_test', 'time', 'this is a description for column time', 'my_catalog');
select ai.sc_set_table_col_desc(42, 9, 2, 'postgres_air', 'hypertable_test', 'location', 'this is a description for column location', 'my_catalog');
select ai.sc_set_table_col_desc(42, 9, 3, 'postgres_air', 'hypertable_test', 'time_received', 'this is a description for column time_received', 'my_catalog');
select ai.sc_set_table_col_desc(42, 9, 4, 'postgres_air', 'hypertable_test', 'params', 'this is a description for column params', 'my_catalog');
select ai.sc_set_table_desc(42, 10, 'postgres_air', 'passenger', 'this is a description for table passenger', 'my_catalog');
select ai.sc_set_table_col_desc(42, 10, 1, 'postgres_air', 'passenger', 'passenger_id', 'this is a description for column passenger_id', 'my_catalog');
select ai.sc_set_table_col_desc(42, 10, 2, 'postgres_air', 'passenger', 'booking_id', 'this is a description for column booking_id', 'my_catalog');
select ai.sc_set_table_col_desc(42, 10, 3, 'postgres_air', 'passenger', 'booking_ref', 'this is a description for column booking_ref', 'my_catalog');
select ai.sc_set_table_col_desc(42, 10, 4, 'postgres_air', 'passenger', 'passenger_no', 'this is a description for column passenger_no', 'my_catalog');
select ai.sc_set_table_col_desc(42, 10, 5, 'postgres_air', 'passenger', 'first_name', 'this is a description for column first_name', 'my_catalog');
select ai.sc_set_table_col_desc(42, 10, 6, 'postgres_air', 'passenger', 'last_name', 'this is a description for column last_name', 'my_catalog');
select ai.sc_set_table_col_desc(42, 10, 7, 'postgres_air', 'passenger', 'account_id', 'this is a description for column account_id', 'my_catalog');
select ai.sc_set_table_col_desc(42, 10, 8, 'postgres_air', 'passenger', 'update_ts', 'this is a description for column update_ts', 'my_catalog');
select ai.sc_set_table_col_desc(42, 10, 9, 'postgres_air', 'passenger', 'age', 'this is a description for column age', 'my_catalog');
select ai.sc_set_table_desc(42, 11, 'postgres_air', 'phone', 'this is a description for table phone', 'my_catalog');
select ai.sc_set_table_col_desc(42, 11, 1, 'postgres_air', 'phone', 'phone_id', 'this is a description for column phone_id', 'my_catalog');
select ai.sc_set_table_col_desc(42, 11, 2, 'postgres_air', 'phone', 'account_id', 'this is a description for column account_id', 'my_catalog');
select ai.sc_set_table_col_desc(42, 11, 3, 'postgres_air', 'phone', 'phone', 'this is a description for column phone', 'my_catalog');
select ai.sc_set_table_col_desc(42, 11, 4, 'postgres_air', 'phone', 'phone_type', 'this is a description for column phone_type', 'my_catalog');
select ai.sc_set_table_col_desc(42, 11, 5, 'postgres_air', 'phone', 'primary_phone', 'this is a description for column primary_phone', 'my_catalog');
select ai.sc_set_table_col_desc(42, 11, 6, 'postgres_air', 'phone', 'update_ts', 'this is a description for column update_ts', 'my_catalog');
select ai.sc_set_view_desc(42, 100, 'postgres_air', 'events_daily', 'this is a description for view events_daily', 'my_catalog');
select ai.sc_set_view_col_desc(42, 100, 1, 'postgres_air', 'events_daily', 'name', 'this is a description for column name', 'my_catalog');
select ai.sc_set_view_col_desc(42, 100, 2, 'postgres_air', 'events_daily', 'bucket', 'this is a description for column bucket', 'my_catalog');
select ai.sc_set_view_desc(42, 101, 'postgres_air', 'flight_summary', 'this is a description for view flight_summary', 'my_catalog');
select ai.sc_set_view_col_desc(42, 101, 1, 'postgres_air', 'flight_summary', 'flight_no', 'this is a description for column flight_no', 'my_catalog');
select ai.sc_set_view_col_desc(42, 101, 2, 'postgres_air', 'flight_summary', 'departure_airport', 'this is a description for column departure_airport', 'my_catalog');
select ai.sc_set_view_col_desc(42, 101, 3, 'postgres_air', 'flight_summary', 'arrival_airport', 'this is a description for column arrival_airport', 'my_catalog');
select ai.sc_set_view_col_desc(42, 101, 4, 'postgres_air', 'flight_summary', 'scheduled_departure', 'this is a description for column scheduled_departure', 'my_catalog');
select ai.sc_set_view_desc(42, 102, 'postgres_air', 'passenger_details', 'this is a description for view passenger_details', 'my_catalog');
select ai.sc_set_view_col_desc(42, 102, 1, 'postgres_air', 'passenger_details', 'first_name', 'this is a description for column first_name', 'my_catalog');
select ai.sc_set_view_col_desc(42, 102, 2, 'postgres_air', 'passenger_details', 'last_name', 'this is a description for column last_name', 'my_catalog');
select ai.sc_set_view_col_desc(42, 102, 3, 'postgres_air', 'passenger_details', 'booking_ref', 'this is a description for column booking_ref', 'my_catalog');
select ai.sc_set_proc_desc(666, 0, 'postgres_air', 'advance_air_time', '{integer,pg_catalog.text,boolean}', 'this is a description for procedure advance_air_time', 'my_catalog');
select ai.sc_set_proc_desc(666, 1, 'postgres_air', 'update_flight_status', '{integer,pg_catalog.text}', 'this is a description for procedure update_flight_status', 'my_catalog');
